# TODO: Module docstring

import abc


class Descriptor(metaclass=abc.ABCMeta):
    # TODO: Docstring
    @abc.abstractmethod
    def __eq__(self, other):
        pass

    @abc.abstractmethod
    def __hash__(self):
        pass



# TODO: Module docstring

from collections.abc import Callable
from dis import dis

from .descriptor import Descriptor


class Predicate(Descriptor):
    """A predicate class.

    A PermSet can be built with a Predicate instance by using the predicate
    provided to it to see if a perm should be in the PermSet or not.
    """

    def __init__(self, predicate):
        if isinstance(predicate, Callable):
            self.predicate = predicate
        else:
            message = "{} object is not callable".format(repr(predicate))
            raise TypeError(message)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            # Disassemble predicates and see if code is the same
            return dis(self.predicate) == dis(other.predicate)
        else:
            return False



import abc

ABC = abc.ABCMeta("ABC", (object,), {})


class Patt(ABC):
    def avoided_by(self, *perms):
        """Check if self is avoided by perms.

        Args:
            self:
                A classical/mesh pattern.
            perms: [permuta.Permutation]
                A list of permutations.

        Returns: bool
            True if and only if every permutation in perms avoids self.
        """
        return all(self not in perm for perm in perms)

    def contained_in(self, *perms):
        """Check if self is a pattern of perms.

        Args:
            self:
                A classical/mesh pattern.
            perms: [permuta.Permutation]
                A list of permutations.

        Returns: bool
            True if and only if self is a pattern of all permutations in perms.
        """
        return all(self in perm for perm in perms)

    def count_occurrences_in(self, perm):
        """Count the number of occurrences of self in perm.

        Args:
            self:
                A classical/mesh pattern.
            perm: permuta.Permutation
                A permutation.

        Returns: int
            The number of times self occurs in perm.
        """
        return sum(1 for _ in self.occurrences_in(perm))

    @abc.abstractmethod
    def occurrences_in(self, perm):
        """Find all indices of occurrences of self in perm."""
        pass



# TODO: Module docstring

import itertools
import random
from math import factorial

from ....perm import Perm
from ...finite.permset_finite import PermSetFinite
from ...finite.permset_finite_specificlength import PermSetFiniteSpecificLength
from ...finite.permset_static import PermSetStatic
from ..permset_unbounded import PermSetUnbounded


class PermSetAll(PermSetUnbounded):
    __iter = None
    __iter_number = None

    def up_to(self, perm):
        # Should return a PermSetAllRange
        raise NotImplementedError

    def of_length(self, length):
        return PermSetAllSpecificLength(length)

    def range(self, stop):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, key):
        return Perm.unrank(key)

    def __next__(self):
        if self.__iter is None:
            self.__iter = iter(self[self.__iter_number])
        try:
            return next(self.__iter)
        except StopIteration:
            self.__iter = None
            self.__iter_number += 1
            return self.__next__()

    def __iter__(self):
        self.__iter = None
        self.__iter_number = 0
        return self

    def __contains__(self, perm):
        return isinstance(perm, Perm)  # Why would you even ask?

    def __repr__(self):
        return "PermSet()"

    def __str__(self):
        return "<The set of all perms>"


class PermSetAllSpecificLength(PermSetFiniteSpecificLength):
    """Class for iterating through all perms of a specific length."""

    __slots__ = "_length"

    def __init__(self, length):
        self._length = length

    @property
    def length(self):
        return self._length

    @property
    def domain(self):
        return list(range(self.length))  # tuple instead?

    def of_length(self, length):
        if length != self._length:
            return PermSetStatic([])
        else:
            return self

    def random(self):
        """Return a random perm of the length."""
        all_elements = self.domain
        random.shuffle(all_elements)
        return Perm(all_elements)

    def __contains__(self, other):
        """Check if other is a permutation in the set."""
        return isinstance(other, Perm) and len(other) == self.length

    def __getitem__(self, key):
        return Perm.unrank(key, self.length)

    def __iter__(self):
        # Need to return new instance because permutations of itertools
        # depletes self
        # This probably needs looking into because the iter isn't a subclass of
        # PermSet
        return PermSetAllSpecificLengthIterator(self.domain)

    def __len__(self):
        return factorial(self.length)

    def __repr__(self):
        return "PermSet({})".format(self.length)

    def __str__(self):
        return "<The set of all perms of length {}>".format(self.length)


class PermSetAllSpecificLengthIterator(itertools.permutations):
    def __next__(self):
        return Perm(super(PermSetAllSpecificLengthIterator, self).__next__())


class PermSetAllFiniteLengthSubset(PermSetFinite):
    pass


class PermSetAllUnboundedLengthSubset(PermSetUnbounded):
    pass



from .meshpatt import MeshPatt
from .perm import Perm
from .permset import PermSet


def gen_meshpatts(length, patt=None):
    """Generates all mesh patterns of length n. If the classical pattern is
    specified then only the mesh patterns with the classical pattern as the
    underlying pattern are generated.

    Args:
        length: <numbers.Integral>
            The length(size) of the mesh pattern.
        patt: <permutation.Permutation>, <numbers.Integral> or
              <collections.Iterable>

    Yields: <permuta.MeshPatt>
        Every permutation of the specified length with the specified classical
        pattern or each of them if the pattern is not specified.

    Examples:
        >>> mps = list(gen_meshpatts(0))
        >>> len(mps)
        2
        >>> mps[0]
        MeshPatt(Perm(()), [])
        >>> mps[1]
        MeshPatt(Perm(()), [(0, 0)])
        >>> len(list(gen_meshpatts(2, (1, 2))))
        512
    """
    if patt is None:
        for p in PermSet(length):
            for i in range(2 ** ((length + 1) ** 2)):
                yield MeshPatt.unrank(p, i)
    else:
        if not isinstance(patt, Perm):
            patt = Perm(patt)
        for i in range(2 ** ((length + 1) ** 2)):
            yield MeshPatt.unrank(patt, i)



class Node:
    def __init__(self, row, col):
        self.left = None
        self.right = None
        self.u = None
        self.d = None
        self.p = None
        self.row = row
        self.col = col
        self.size = 0


def cover(c):
    c.right.left = c.left
    c.left.right = c.right
    i = c.d
    while i != c:
        j = i.right
        while j != i:
            j.d.u = j.u
            j.u.d = j.d
            j.p.size -= 1
            j = j.right
        i = i.d


def uncover(c):
    i = c.u
    while i != c:
        j = i.left
        while j != i:
            j.p.size += 1
            j.u.d = j
            j.d.u = j.u.d
            j = j.left
        i = i.u
    c.left.right = c
    c.right.left = c.left.right


class AlgorithmX:
    def __init__(self, rows, cols, solution_callback):
        assert rows > 0 and cols > 0
        self.rows = rows
        self.cols = cols
        self.head = None
        self.arr = [[False for c in range(self.cols)] for r in range(self.rows)]
        self.sol = [0 for i in range(self.rows)]
        self.solution_callback = solution_callback
        self.can_continue = False

    def set_value(self, row, col, val=True):
        self.arr[row][col] = val

    def setup(self):
        ptr = [
            [
                Node(i, j) if i == self.rows or self.arr[i][j] else None
                for j in range(self.cols)
            ]
            for i in range(self.rows + 1)
        ]
        for i in range(self.rows + 1):
            for j in range(self.cols):
                if ptr[i][j] is None:
                    continue

                ni = i + 1
                nj = j + 1

                while True:
                    if ni == self.rows + 1:
                        ni = 0
                    if ni == self.rows or self.arr[ni][j]:
                        break
                    ni += 1

                ptr[i][j].d = ptr[ni][j]
                ptr[ni][j].u = ptr[i][j]

                while True:
                    if nj == self.cols:
                        nj = 0
                    if i == self.rows or self.arr[i][nj]:
                        break
                    nj += 1

                ptr[i][j].right = ptr[i][nj]
                ptr[i][nj].left = ptr[i][j]

        self.head = Node(self.rows, -1)
        self.head.right = ptr[self.rows][0]
        ptr[self.rows][0].left = self.head
        self.head.left = ptr[self.rows][self.cols - 1]
        ptr[self.rows][self.cols - 1].right = self.head

        for j in range(self.cols):
            cnt = -1
            for i in range(self.rows + 1):
                if ptr[i][j] is not None:
                    cnt += 1
                    ptr[i][j].p = ptr[self.rows][j]
            ptr[self.rows][j].size = cnt

    def search(self, k=0, at_most=None):
        if self.head == self.head.right:
            res = [self.sol[i] for i in range(k)]
            res = sorted(res)
            return self.solution_callback(res)

        if at_most is not None and k >= at_most:
            self.can_continue = True
            return

        c = self.head.right
        tmp = self.head.right
        while tmp != self.head:
            if tmp.size < c.size:
                c = tmp
            tmp = tmp.right

        if c == c.d:
            return False

        cover(c)

        found = False
        r = c.d
        while not found and r != c:
            self.sol[k] = r.row

            j = r.right
            while j != r:
                cover(j.p)
                j = j.right

            found = self.search(k + 1, at_most)

            j = r.left
            while j != r:
                uncover(j.p)
                j = j.left

            r = r.d

        uncover(c)
        return found



import itertools
import random

import pytest
from permuta import MeshPatt, Perm
from permuta.meshpattset import gen_meshpatts
from permuta.misc import DIR_EAST, DIR_NORTH, DIR_SOUTH, DIR_WEST, factorial

mesh_pattern = MeshPatt(
    [1, 3, 2, 0], set([(0, 0), (4, 0), (2, 1), (4, 1), (2, 2), (4, 2), (3, 3), (0, 4)])
)
perm1 = Perm([5, 2, 8, 6, 7, 9, 4, 3, 1, 0])  # Occurrence: E.g., [1, 3, 6, 9]
perm2 = Perm([1, 2, 8, 6, 5, 9, 4, 3, 7, 0])  # Occurrence: E.g., [1, 6, 7, 9]
perm3 = Perm([9, 8, 7, 6, 2, 5, 3, 4, 0, 1])  # Occurrence: None (avoids)
perm4 = Perm([0, 1, 2, 3, 4])  # Avoids as well
perm5 = Perm([1, 2, 4, 3, 0])  # Two occurrences
patt1 = perm4
patt2 = perm5
patt3 = Perm((2, 3, 0, 1))
shad1 = frozenset(
    [(0, 0), (1, 0), (2, 0), (2, 1), (3, 2), (3, 3), (5, 0), (5, 1), (5, 2)]
)
shad2 = frozenset([(3, 3), (2, 2), (1, 1), (0, 2)])
shad3 = frozenset([(1, 3), (4, 4), (2, 1), (2, 2), (0, 4), (4, 0)])
mesh1 = MeshPatt(patt1, shad1)
mesh2 = MeshPatt(patt2, shad2)
mesh3 = MeshPatt(patt3, shad3)


def test_init():
    with pytest.raises(ValueError):
        MeshPatt(Perm([0, 1, 1], check=True), ())
    with pytest.raises(ValueError):
        MeshPatt(Perm([1, 0, 1], check=True), ())
    with pytest.raises(ValueError):
        MeshPatt(Perm([0, 0], check=True), ())
    with pytest.raises(ValueError):
        MeshPatt(Perm([1], check=True), ())
    with pytest.raises(ValueError):
        MeshPatt(Perm((1,), check=True), ())
    with pytest.raises(TypeError):
        MeshPatt(Perm(101, check=True), ())
    with pytest.raises(TypeError):
        MeshPatt(Perm(-234, check=True), ())
    with pytest.raises(TypeError):
        MeshPatt(Perm(None, check=True), ())
    with pytest.raises(TypeError):
        MeshPatt(Perm([0.1, 0.2, 0.3], check=True), ())
    with pytest.raises(TypeError):
        MeshPatt(Perm((), check=True), (0, 1))
    with pytest.raises(TypeError):
        MeshPatt(Perm((0, 1, 2), check=True), [(1, "a")])
    with pytest.raises(TypeError):
        MeshPatt(Perm((0, 1, 2), check=True), [("a", 1)])
    with pytest.raises(ValueError):
        MeshPatt(Perm.random(5), [(0,), (1, 1)])
    with pytest.raises(ValueError):
        MeshPatt(Perm.random(5), [(0, 0, 0), (1, 1, 1)])
    with pytest.raises(ValueError):
        MeshPatt(Perm((), check=True), [(0, 1), (1, 0)])
    with pytest.raises(ValueError):
        MeshPatt(Perm.random(3), [(0, -1), (0, 0)])
    with pytest.raises(ValueError):
        MeshPatt(Perm.random(10), [(0, 0), (12, 7)])
    MeshPatt(Perm(check=True), ())
    MeshPatt(Perm([], check=True), ())
    MeshPatt(Perm((0,), check=True), ())
    MeshPatt(Perm([0], check=True), ())
    MeshPatt(Perm([3, 0, 2, 1], check=True), ())
    MeshPatt(
        Perm([3, 0, 2, 1], check=True),
        [
            (0, 0),
            (0, 1),
            (0, 2),
            (0, 3),
            (0, 4),
            (1, 0),
            (1, 1),
            (1, 2),
            (1, 3),
            (1, 4),
            (2, 0),
            (2, 1),
            (2, 2),
            (2, 3),
            (2, 4),
            (3, 0),
            (3, 1),
            (3, 2),
            (3, 3),
            (3, 4),
            (4, 0),
            (4, 1),
            (4, 2),
            (4, 3),
            (4, 4),
        ],
    )
    MeshPatt(Perm([3, 0, 2, 1], check=True), [(0, 0), (0, 1), (0, 2), (0, 3), (0, 4)])
    MeshPatt([3, 0, 2, 1], [(0, 2), (0, 3), (0, 4)])
    MeshPatt(set([3, 0, 2, 1]), [(0, 2), (0, 3), (0, 4)])


def test_complement():
    assert MeshPatt(Perm(), []).complement() == MeshPatt(Perm(), [])
    assert MeshPatt(Perm((0,)), []).complement() == MeshPatt(Perm((0,)), [])
    assert MeshPatt(Perm((0,)), [(0, 0)]).complement() == MeshPatt(Perm((0,)), [(0, 1)])
    assert MeshPatt(Perm((0,)), [(0, 1)]).complement() == MeshPatt(Perm((0,)), [(0, 0)])
    assert MeshPatt(Perm((0,)), [(1, 0)]).complement() == MeshPatt(Perm((0,)), [(1, 1)])
    assert MeshPatt(Perm((0,)), [(1, 1)]).complement() == MeshPatt(Perm((0,)), [(1, 0)])
    for _ in range(20):
        mpatt = MeshPatt.random(10)
        comp = mpatt.complement()
        assert len(mpatt.pattern) == len(comp.pattern)
        assert comp.complement() == mpatt


def test_flip_horizontal():
    assert MeshPatt(Perm(), []).flip_vertical() == MeshPatt(Perm(), [])
    assert MeshPatt(Perm((0,)), [(0, 1)]).flip_horizontal() == MeshPatt(
        Perm((0,)), [(0, 0)]
    )


def test_reverse():
    assert MeshPatt(Perm(), []).reverse() == MeshPatt(Perm(), [])
    assert MeshPatt(Perm((0,)), []).reverse() == MeshPatt(Perm((0,)), [])
    assert MeshPatt(Perm((0,)), [(0, 0)]).reverse() == MeshPatt(Perm((0,)), [(1, 0)])
    assert MeshPatt(Perm((0,)), [(0, 1)]).reverse() == MeshPatt(Perm((0,)), [(1, 1)])
    assert MeshPatt(Perm((0,)), [(1, 0)]).reverse() == MeshPatt(Perm((0,)), [(0, 0)])
    assert MeshPatt(Perm((0,)), [(1, 1)]).reverse() == MeshPatt(Perm((0,)), [(0, 1)])
    for _ in range(20):
        mpatt = MeshPatt.random(10)
        comp = mpatt.reverse()
        assert len(mpatt.pattern) == len(comp.pattern)
        assert comp.reverse() == mpatt


def test_flip_vertical():
    assert MeshPatt(Perm(), []).flip_vertical() == MeshPatt(Perm(), [])
    assert MeshPatt(Perm((0,)), [(0, 0)]).flip_vertical() == MeshPatt(
        Perm((0,)), [(1, 0)]
    )


def test_inverse():
    assert MeshPatt(Perm(), []).inverse() == MeshPatt(Perm(), [])
    assert MeshPatt(Perm((0,)), []).inverse() == MeshPatt(Perm((0,)), [])
    assert MeshPatt(Perm((0,)), [(0, 0)]).inverse() == MeshPatt(Perm((0,)), [(0, 0)])
    assert MeshPatt(Perm((0,)), [(0, 1)]).inverse() == MeshPatt(Perm((0,)), [(1, 0)])
    assert MeshPatt(Perm((0,)), [(1, 0)]).inverse() == MeshPatt(Perm((0,)), [(0, 1)])
    assert MeshPatt(Perm((0,)), [(1, 1)]).inverse() == MeshPatt(Perm((0,)), [(1, 1)])
    for _ in range(20):
        mpatt = MeshPatt.random(10)
        comp = mpatt.inverse()
        assert len(mpatt.pattern) == len(comp.pattern)
        assert comp.inverse() == mpatt


def test_sub_mesh_pattern():
    # Empty pattern
    assert mesh1.sub_mesh_pattern(()) == MeshPatt((), ())
    # Sub mesh pattern from indices 1, 2, and 3 of mesh1
    pattern = (0, 1, 2)
    shading = set([(1, 0), (2, 1), (2, 2)])
    mesh_pattern = MeshPatt(pattern, shading)
    sub_mesh_pattern = mesh1.sub_mesh_pattern((1, 2, 3))
    assert sub_mesh_pattern == mesh_pattern
    # Sub mesh pattern from indices 0, 1, and 4 of mesh1
    pattern = (0, 1, 2)
    shading = set([(0, 0), (1, 0), (3, 0), (3, 1)])
    mesh_pattern = MeshPatt(pattern, shading)
    sub_mesh_pattern = mesh1.sub_mesh_pattern((0, 1, 4))
    assert sub_mesh_pattern == mesh_pattern
    # Sub mesh pattern from indices 3 and 4 of mesh1
    assert mesh1.sub_mesh_pattern((3, 4)) == MeshPatt((0, 1))
    # Sub mesh pattern from indices 2 and 3 of mesh1
    assert mesh1.sub_mesh_pattern((2, 3)) == MeshPatt((0, 1), set([(1, 1)]))
    # Sub mesh pattern from index 0 of mesh3
    assert mesh3.sub_mesh_pattern((0,)) == MeshPatt((0,))
    # Sub mesh pattern from indices 0, 1, and 3 of mesh3
    assert mesh3.sub_mesh_pattern((0, 1, 3)) == MeshPatt(
        (1, 2, 0), set([(0, 3), (1, 2), (3, 3)])
    )
    # Some complete sub meshes
    assert mesh1.sub_mesh_pattern(range(len(mesh1))) == mesh1
    assert mesh2.sub_mesh_pattern(range(len(mesh2))) == mesh2
    assert mesh3.sub_mesh_pattern(range(len(mesh3))) == mesh3

    mpatt = MeshPatt((0, 1), [(0, 0), (0, 1), (1, 0), (1, 1)])
    assert mpatt.sub_mesh_pattern([0]) == MeshPatt((0,), [(0, 0)])
    assert mpatt.sub_mesh_pattern([1]) == MeshPatt((0,))

    mpatt = MeshPatt((0, 1, 2), [(0, 1), (1, 1), (1, 2), (2, 1), (2, 2), (3, 1)])
    mpatt.sub_mesh_pattern([0, 1]) == MeshPatt((0, 1), [(0, 1), (1, 1), (2, 1)])
    mpatt.sub_mesh_pattern([0, 2]) == MeshPatt((0, 1), [])


def test_flip_diagonal():
    assert MeshPatt(Perm(), []).flip_diagonal() == MeshPatt(Perm(), [])
    assert MeshPatt(Perm((0,)), [(0, 1)]).flip_diagonal() == MeshPatt(
        Perm((0,)), [(1, 0)]
    )


def test_rotate():
    assert MeshPatt(Perm(), [])._rotate_right() == MeshPatt(Perm(), [])
    assert MeshPatt(Perm(), [])._rotate_left() == MeshPatt(Perm(), [])
    assert MeshPatt(Perm(), [])._rotate_180() == MeshPatt(Perm(), [])

    assert MeshPatt(Perm((0,)), [])._rotate_right() == MeshPatt(Perm((0,)), [])
    assert MeshPatt(Perm((0,)), [])._rotate_left() == MeshPatt(Perm((0,)), [])
    assert MeshPatt(Perm((0,)), [])._rotate_180() == MeshPatt(Perm((0,)), [])

    for _ in range(50):
        mpatt = MeshPatt.random(6)
        assert mpatt._rotate_right()._rotate_right() == mpatt._rotate_180()
        assert mpatt._rotate_left()._rotate_left() == mpatt._rotate_180()
        assert mpatt._rotate_right()._rotate_left() == mpatt
        assert mpatt._rotate_left()._rotate_right() == mpatt
        assert mpatt._rotate_180()._rotate_180() == mpatt


def test_shade():
    assert MeshPatt().shade((0, 0)).is_shaded((0, 0))
    assert MeshPatt().shade([(0, 0)]).is_shaded((0, 0))

    newshad = [(1, 2), (3, 3), (4, 4)]
    mesh1shaded = mesh1.shade(newshad)
    for shading in shad1:
        assert mesh1shaded.is_shaded(shading)
    for shading in newshad:
        assert mesh1shaded.is_shaded((shading))

    newshad = [(2, 2), (1, 1), (0, 2), (1, 2), (3, 3), (4, 4)]
    mesh2shaded = mesh2.shade(newshad)
    for shading in shad2:
        assert mesh2shaded.is_shaded(shading)
    for shading in newshad:
        assert mesh2shaded.is_shaded(shading)

    with pytest.raises(ValueError):
        mesh2.shade(())
    with pytest.raises(ValueError):
        mesh1.shade(())


def test_add_point():
    mpatt = MeshPatt()
    assert mpatt.add_point((0, 0)) == MeshPatt((0,))
    assert mpatt.add_point((0, 0), shade_dir=DIR_EAST) == MeshPatt(
        (0,), [(1, 0), (1, 1)]
    )
    assert mpatt.add_point((0, 0), shade_dir=DIR_NORTH) == MeshPatt(
        (0,), [(0, 1), (1, 1)]
    )
    assert mpatt.add_point((0, 0), shade_dir=DIR_WEST) == MeshPatt(
        (0,), [(0, 0), (0, 1)]
    )
    assert mpatt.add_point((0, 0), shade_dir=DIR_SOUTH) == MeshPatt(
        (0,), [(0, 0), (1, 0)]
    )

    mpatt = MeshPatt((0, 1, 2), [(1, 0), (2, 1), (3, 2)])
    assert mpatt.add_point((2, 0), shade_dir=DIR_SOUTH) == MeshPatt(
        (1, 2, 0, 3), [(1, 0), (1, 1), (2, 0), (2, 2), (3, 0), (3, 2), (4, 3)]
    )
    assert mpatt.add_point((1, 2), shade_dir=DIR_WEST) == MeshPatt(
        (0, 2, 1, 3), [(1, 0), (1, 2), (1, 3), (2, 0), (3, 1), (4, 2), (4, 3)]
    )
    assert mesh1.add_point((2, 3), shade_dir=DIR_NORTH) == MeshPatt(
        (0, 1, 3, 2, 4, 5),
        [
            (0, 0),
            (1, 0),
            (2, 0),
            (2, 1),
            (2, 4),
            (3, 0),
            (3, 1),
            (3, 4),
            (4, 2),
            (4, 3),
            (4, 4),
            (6, 0),
            (6, 1),
            (6, 2),
        ],
    )

    with pytest.raises(TypeError):
        mpatt.add_point(("a", (0,)))
    with pytest.raises(ValueError):
        mpatt.add_point((2, 1))


def test_add_increase():
    mpatt = MeshPatt()
    assert mpatt.add_increase((0, 0)) == MeshPatt((0, 1))
    mpatt = MeshPatt((0, 1, 2), [(1, 0), (2, 1), (3, 2)])
    assert mpatt.add_increase((2, 0)) == MeshPatt(
        Perm((2, 3, 0, 1, 4)), [(1, 2), (5, 4), (3, 3), (2, 3), (4, 3), (1, 0), (1, 1)]
    )
    assert mpatt.add_increase((1, 2)) == MeshPatt(
        (0, 2, 3, 1, 4), [(1, 0), (2, 0), (3, 0), (4, 1), (5, 2), (5, 3), (5, 4)]
    )


def test_add_decrease():
    mpatt = MeshPatt()
    assert mpatt.add_decrease((0, 0)) == MeshPatt((1, 0))
    mpatt = MeshPatt((0, 1, 2), [(1, 0), (2, 1), (3, 2)])
    assert mpatt.add_decrease((2, 0)) == MeshPatt(
        Perm((2, 3, 1, 0, 4)), [(1, 2), (5, 4), (3, 3), (2, 3), (4, 3), (1, 0), (1, 1)]
    )
    assert mpatt.add_decrease((1, 2)) == MeshPatt(
        (0, 3, 2, 1, 4), [(1, 0), (2, 0), (3, 0), (4, 1), (5, 2), (5, 3), (5, 4)]
    )


def test_contained_in():
    assert Perm([0, 1, 2]).contains(MeshPatt([0, 1], set([(1, 0), (1, 1), (1, 2)])))
    assert mesh_pattern.contained_in(perm1)
    assert mesh_pattern.contained_in(perm2)
    assert not (mesh_pattern.contained_in(perm3))
    assert mesh_pattern.contained_in(perm1, perm2)
    assert not (mesh_pattern.contained_in(perm1, perm3))


def test_meshpatt_contains_meshpatt():
    p_12 = Perm([0, 1])
    p_123 = Perm([0, 1, 2])
    small_mesh_patt = MeshPatt(p_12, [(1, 0), (1, 1), (1, 2)])
    big_mesh_patt_1st_col = MeshPatt(p_123, [(1, 0), (1, 1), (1, 2), (1, 3)])
    big_mesh_patt_2nd_col = MeshPatt(p_123, [(2, 0), (2, 1), (2, 2), (2, 3)])
    assert small_mesh_patt in big_mesh_patt_1st_col
    assert small_mesh_patt in big_mesh_patt_2nd_col
    assert big_mesh_patt_1st_col not in small_mesh_patt
    assert big_mesh_patt_2nd_col not in small_mesh_patt
    assert big_mesh_patt_1st_col not in big_mesh_patt_2nd_col
    assert big_mesh_patt_2nd_col not in big_mesh_patt_1st_col


def test_avoided_by():
    assert not (mesh_pattern.avoided_by(perm1))
    assert not (mesh_pattern.avoided_by(perm2))
    assert mesh_pattern.avoided_by(perm3)
    assert mesh_pattern.avoided_by(perm4)
    assert mesh_pattern.avoided_by(perm3, perm4)
    assert not (mesh_pattern.avoided_by(perm4, perm1, perm3))


def test_count_occurrences_in():
    assert mesh_pattern.count_occurrences_in(perm1) == 8
    assert mesh_pattern.count_occurrences_in(perm2) == 12
    assert mesh_pattern.count_occurrences_in(perm3) == 0
    assert mesh_pattern.count_occurrences_in(perm4) == 0
    assert mesh_pattern.count_occurrences_in(perm5) == 2


def test_is_shaded():
    mpatt = MeshPatt(Perm(), ((0, 0),))
    assert mpatt.is_shaded((0, 0))
    mpatt = MeshPatt(
        (0, 2, 1), [(0, 0), (0, 2), (1, 1), (1, 3), (2, 0), (2, 2), (3, 1), (3, 3)]
    )
    for (x0, y0) in itertools.combinations(range(len(mpatt) + 1), 2):
        for (x1, y1) in itertools.combinations(range(len(mpatt) + 1), 2):
            if x0 > x1 or y0 > y1:
                with pytest.raises(ValueError):
                    mpatt.is_shaded((x0, y0), (x1, y1))
            elif x0 == x1 and y0 == y1:
                if (x0 + y0) % 2 == 0:
                    assert mpatt.is_shaded((x0, y0))
                    assert mpatt.is_shaded((x0, y0), (x1, y1))
            else:
                assert not mpatt.is_shaded((x0, y0), (x1, y1))
    with pytest.raises(ValueError):
        mpatt.is_shaded((4, 0))
    with pytest.raises(ValueError):
        mpatt.is_shaded((0, 4))
    with pytest.raises(ValueError):
        mpatt.is_shaded((-1, 2))
    with pytest.raises(ValueError):
        mpatt.is_shaded((0, 0), (0, 4))


def test_is_pointfree():
    mpatt = MeshPatt((0, 1), [(0, 0), (0, 1), (1, 0), (1, 1)])
    assert not mpatt.is_pointfree((0, 0), (1, 1))
    assert not mpatt.is_pointfree((1, 1), (2, 2))
    assert not mpatt.is_pointfree((0, 1), (2, 2))
    assert mpatt.is_pointfree((0, 1), (2, 1))
    assert not mpatt.is_pointfree((0, 1), (2, 2))
    mpatt = MeshPatt((0, 1, 2), [(0, 1), (1, 1), (1, 2), (2, 1), (2, 2), (3, 1)])
    assert not mpatt.is_pointfree((1, 1), (2, 2))


def test_can_shade():
    assert not (MeshPatt().can_shade((0, 0)))
    assert MeshPatt((0,)).can_shade((0, 0)) == [0]
    assert MeshPatt((0,)).can_shade((1, 0)) == [0]
    assert not (MeshPatt((0,), [(0, 0)]).can_shade((0, 0)))
    assert not (MeshPatt((0,), [(0, 0)]).can_shade((1, 1)))
    assert not mesh_pattern.can_shade((1, 1))
    assert not mesh_pattern.can_shade((3, 2))
    assert not mesh_pattern.can_shade((1, 2))
    assert mesh_pattern.can_shade((0, 1)) == [1]
    mpatt = MeshPatt((1, 2, 0), [(2, 2), (3, 0), (3, 2), (3, 3)])
    assert list(sorted(mpatt.can_shade((1, 2)))) == [1, 2]
    assert mpatt.can_shade((3, 1)) == [0]


def test_can_simul_shade():
    assert not (MeshPatt().can_simul_shade((0, 0), (0, 0)))
    assert MeshPatt((0,)).can_simul_shade((0, 0), (0, 1)) == [0]
    assert not (MeshPatt((0, 1, 2)).can_simul_shade((2, 0), (2, 2)))
    assert not mesh1.can_simul_shade((2, 2), (3, 2))
    assert not mesh1.can_simul_shade((1, 2), (2, 2))
    assert not mesh2.can_simul_shade((3, 3), (4, 3))
    assert not mesh2.can_simul_shade((3, 4), (4, 4))
    assert not mesh1.can_simul_shade((4, 4), (4, 5))
    assert not mesh1.can_simul_shade((4, 5), (5, 5))
    assert mesh1.can_simul_shade((5, 4), (5, 5)) == [4]

    mpatt = MeshPatt(
        (2, 3, 0, 1), [(0, 4), (1, 3), (2, 1), (2, 2), (3, 4), (4, 0), (4, 4)]
    )
    assert mpatt.can_simul_shade((4, 1), (4, 2)) == [1]
    mpatt = MeshPatt(
        (1, 2, 0), [(0, 2), (0, 3), (1, 1), (2, 0), (2, 1), (3, 2), (3, 3)]
    )
    assert mpatt.can_simul_shade((2, 2), (2, 3)) == [2]
    assert not mpatt.can_simul_shade((1, 2), (2, 2))
    assert not mpatt.can_simul_shade((1, 2), (1, 3))


def test_shadable_boxes():
    assert not (MeshPatt().shadable_boxes())
    assert len(MeshPatt((0,)).shadable_boxes()[0]) == 8
    assert len(MeshPatt((0, 1)).shadable_boxes()[0]) == 8
    assert len(MeshPatt((0, 1)).shadable_boxes()[1]) == 8


def test_non_pointless_boxes():
    assert MeshPatt(Perm()).non_pointless_boxes() == set()
    assert MeshPatt((0,)).non_pointless_boxes() == set([(0, 0), (0, 1), (1, 0), (1, 1)])
    assert MeshPatt(
        (0, 1, 2), [(0, 0), (1, 2,), (0, 2), (2, 0), (2, 1)]
    ).non_pointless_boxes() == set(
        [(0, 0), (0, 1), (1, 0), (1, 1), (1, 2), (2, 1), (2, 2), (2, 3), (3, 2), (3, 3)]
    )
    print(mesh2)
    assert mesh2.non_pointless_boxes() == set(
        [
            (0, 1),
            (0, 2),
            (1, 1),
            (1, 2),
            (1, 3),
            (2, 2),
            (2, 3),
            (2, 4),
            (2, 5),
            (3, 3),
            (3, 4),
            (3, 5),
            (4, 0),
            (4, 1),
            (4, 3),
            (4, 4),
            (5, 0),
            (5, 1),
        ]
    )


def test_has_anchered_point():
    assert not any(MeshPatt(Perm()).has_anchored_point())
    assert all(MeshPatt(Perm(), [(0, 0)]).has_anchored_point())
    assert not any(MeshPatt(Perm.random(5)).has_anchored_point())
    assert all(MeshPatt.unrank((1, 2, 0), 65535))
    right, top, left, bottom = MeshPatt(
        (0, 1, 2), [(0, i) for i in range(4)]
    ).has_anchored_point()
    assert left and not (right or top or bottom)
    right, top, left, bottom = MeshPatt(
        (0, 1, 2), [(3, i) for i in range(4)]
    ).has_anchored_point()
    assert right and not (left or top or bottom)
    right, top, left, bottom = MeshPatt(
        (0, 1, 2), [(i, 0) for i in range(4)]
    ).has_anchored_point()
    assert bottom and not (left or top or right)
    right, top, left, bottom = MeshPatt(
        (0, 1, 2), [(i, 3) for i in range(4)]
    ).has_anchored_point()
    assert top and not (left or bottom or right)

    mpatt = MeshPatt.random(5)
    leftshad = [(0, i) for i in range(len(mpatt) + 1)]
    rightshad = [(len(mpatt), i) for i in range(len(mpatt) + 1)]
    upshad = [(i, len(mpatt)) for i in range(len(mpatt) + 1)]
    bottomshad = [(i, 0) for i in range(len(mpatt) + 1)]
    right, top, left, bottom = mpatt.shade(leftshad).has_anchored_point()
    assert left
    right, top, left, bottom = mpatt.shade(rightshad).has_anchored_point()
    assert right
    right, top, left, bottom = mpatt.shade(upshad).has_anchored_point()
    assert top
    right, top, left, bottom = mpatt.shade(bottomshad).has_anchored_point()
    assert bottom


def test_rank():
    assert (
        MeshPatt(Perm((0, 1)), [(0, 1), (1, 2), (2, 1), (2, 0), (2, 2), (1, 1)]).rank()
        == 498
    )
    assert (
        MeshPatt(
            Perm((1, 0, 2)),
            [(3, 2), (0, 0), (2, 3), (1, 0), (0, 1), (1, 2), (3, 3), (3, 1), (2, 0)],
        ).rank()
        == 59731
    )
    assert (
        MeshPatt(
            Perm((0, 3, 2, 1)), [(0, 1), (4, 4), (1, 4), (2, 3), (4, 2), (4, 1), (0, 2)]
        ).rank()
        == 23077382
    )

    pattern = Perm([1, 2])
    mesh = MeshPatt(pattern, set([(0, 1), (0, 2), (1, 2), (2, 0), (2, 1)]))
    assert mesh.rank() == 230

    pattern = Perm([3, 1, 2])
    mesh = MeshPatt(
        pattern,
        set(
            [
                (0, 0),
                (0, 1),
                (0, 2),
                (1, 0),
                (1, 2),
                (2, 0),
                (2, 1),
                (2, 2),
                (2, 3),
                (3, 0),
            ]
        ),
    )
    assert mesh.rank() == 8023

    assert mesh1.rank() == 0b111000000001100000011000001000001

    for _ in range(50):
        mesh = MeshPatt.random(random.randint(0, 20))
        assert MeshPatt.unrank(mesh.pattern, mesh.rank()) == mesh


def test_unrank():
    assert MeshPatt.unrank(Perm((0, 1)), 498) == MeshPatt(
        Perm((0, 1)), frozenset({(0, 1), (1, 2), (2, 1), (2, 0), (2, 2), (1, 1)})
    )
    assert MeshPatt.unrank(Perm((1, 0, 2)), 59731) == MeshPatt(
        Perm((1, 0, 2)),
        frozenset(
            {(3, 2), (0, 0), (2, 3), (1, 0), (0, 1), (1, 2), (3, 3), (3, 1), (2, 0)}
        ),
    )
    assert MeshPatt.unrank(Perm((0, 3, 2, 1)), 23077382) == MeshPatt(
        Perm((0, 3, 2, 1)),
        frozenset({(0, 1), (4, 4), (1, 4), (2, 3), (4, 2), (4, 1), (0, 2)}),
    )

    pattern = Perm([1, 2])
    mesh = MeshPatt(pattern, set([(0, 1), (0, 2), (1, 2), (2, 0), (2, 1)]))
    assert MeshPatt.unrank(pattern, 230) == mesh

    pattern = Perm([3, 1, 2])
    mesh = MeshPatt(
        pattern,
        set(
            [
                (0, 0),
                (0, 1),
                (0, 2),
                (1, 0),
                (1, 2),
                (2, 0),
                (2, 1),
                (2, 2),
                (2, 3),
                (3, 0),
            ]
        ),
    )
    assert MeshPatt.unrank(pattern, 8023) == mesh

    assert MeshPatt.unrank(patt1, 0b111000000001100000011000001000001) == mesh1

    for length in range(20):
        m = MeshPatt.unrank(Perm.random(length), 0)
        assert not m.shading
    for length in range(20):
        m = MeshPatt.unrank(Perm.random(length), 2 ** ((length + 1) ** 2) - 1)
        assert len(m.shading) == (length + 1) ** 2

    with pytest.raises(ValueError):
        MeshPatt.unrank([1, 2, 3], -1)
    with pytest.raises(ValueError):
        MeshPatt.unrank(Perm([1]), 16)
    with pytest.raises(ValueError):
        MeshPatt.unrank(Perm([1]), -1)
    with pytest.raises(TypeError):
        MeshPatt.unrank(Perm.random(length), "haha")
    with pytest.raises(TypeError):
        MeshPatt.unrank(Perm([1]), "1")
    with pytest.raises(ValueError):
        MeshPatt.unrank(Perm.random(10), 2 ** ((10 + 1) ** 2) + 1)


def test_random():
    assert MeshPatt.random(0) in set(
        [MeshPatt(Perm(()), []), MeshPatt(Perm(()), [(0, 0)])]
    )
    assert MeshPatt.random(1) in set(
        MeshPatt.unrank(Perm((0,)), i) for i in range(0, 16)
    )
    assert MeshPatt.random(2) in (
        set(MeshPatt.unrank(Perm((0, 1)), i) for i in range(0, 512))
        | set(MeshPatt.unrank(Perm((1, 0)), i) for i in range(0, 512))
    )
    for length in range(3, 20):
        assert len(MeshPatt.random(length)) == length


def test_len():
    assert len(mesh1) == 5
    assert len(mesh2) == 5
    assert len(mesh3) == 4
    assert len(MeshPatt()) == 0
    assert len(MeshPatt((1,))) == 1
    for _ in range(20):
        length = random.randint(0, 20)
        m = MeshPatt(
            Perm.random(length),
            random.sample(
                [(i, j) for i in range(length + 1) for j in range(length + 1)], k=length
            ),
        )
        assert len(m) == length


def test_bool():
    assert mesh1
    assert mesh2
    assert mesh3
    assert MeshPatt(Perm((1,)))
    assert MeshPatt((), set([(0, 0)]))
    assert MeshPatt(Perm([0, 1, 2, 3]), ())
    assert MeshPatt(Perm([0, 1, 2, 3]), ((0, 0),))
    assert MeshPatt(Perm([0]), ())
    assert not (MeshPatt(Perm([]), ()))
    assert not (MeshPatt(Perm(), ()))


def test_eq():
    assert not (mesh1 == mesh2)
    assert not (mesh1 == mesh3)
    assert not (mesh2 == mesh1)
    assert not (mesh2 == mesh3)
    assert not (mesh3 == mesh1)
    assert not (mesh3 == mesh2)
    mesh1copy = MeshPatt(patt1, shad1)
    mesh2copy = MeshPatt(patt2, shad2)
    mesh3copy = MeshPatt(patt3, shad3)
    assert mesh1 == mesh1copy
    assert mesh2 == mesh2copy
    assert mesh3 == mesh3copy
    assert mesh1copy == mesh1copy
    assert mesh2copy == mesh2copy
    assert mesh3copy == mesh3copy


def test_ordering():
    # mesh1, mesh2 and mesh3 have different underlying patterns
    mp1 = MeshPatt(perm1, shad1)
    mp2 = MeshPatt(perm1, shad2)
    mp3 = MeshPatt(perm1, shad3)

    # Ensure some ordering of mesh patts (all should be comparable)
    assert mesh1 > mesh2 or mesh1 <= mesh2
    assert mp1 < mp2 or mp1 >= mp2

    # Describing lexicographical ordering of mesh patts:
    #  1) order first by underlying patts
    #  2) order second by shading [Python list ordering by smallest indices]
    assert mesh3 < mesh1 < mesh2  # case 1)
    assert mp1 < mp2 < mp3  # case 2)


def test_to_tikz():
    assert (
        mesh_pattern.to_tikz()
        == "\\begin{tikzpicture}[scale=.3,baseline=(current bounding box.center)]\n"
        "\t\\foreach \\x in {1,...,4} {\n"
        "\t\t\\draw[ultra thin] (\\x,0)--(\\x,5); %vline\n"
        "\t\t\\draw[ultra thin] (0,\\x)--(5,\\x); %hline\n"
        "\t\n"
        "\t}\n"
        "\t\\fill[pattern color = black!75, pattern=north east lines] (0, 0) rectangle +(1,1);\n"  # noqa: E501
        "\t\\fill[pattern color = black!75, pattern=north east lines] (0, 4) rectangle +(1,1);\n"  # noqa: E501
        "\t\\fill[pattern color = black!75, pattern=north east lines] (2, 1) rectangle +(1,1);\n"  # noqa: E501
        "\t\\fill[pattern color = black!75, pattern=north east lines] (2, 2) rectangle +(1,1);\n"  # noqa: E501
        "\t\\fill[pattern color = black!75, pattern=north east lines] (3, 3) rectangle +(1,1);\n"  # noqa: E501
        "\t\\fill[pattern color = black!75, pattern=north east lines] (4, 0) rectangle +(1,1);\n"  # noqa: E501
        "\t\\fill[pattern color = black!75, pattern=north east lines] (4, 1) rectangle +(1,1);\n"  # noqa: E501
        "\t\\fill[pattern color = black!75, pattern=north east lines] (4, 2) rectangle +(1,1);\n"  # noqa: E501
        "\t\\draw[fill=black] (1,2) circle (5pt);\n"
        "\t\\draw[fill=black] (2,4) circle (5pt);\n"
        "\t\\draw[fill=black] (3,3) circle (5pt);\n"
        "\t\\draw[fill=black] (4,1) circle (5pt);\n"
        "\\end{tikzpicture}"
    )


def test_ascii_plot():
    assert (
        mesh_pattern.ascii_plot()
        == "▒| | | |\n-+-●-+-+-\n | | |▒|\n-+-+-●-+-\n | |▒| |▒\n-●-+-+-+-\n"
        " | |▒| |▒\n-+-+-+-●-\n▒| | | |▒"
    )


def test_gen_meshpatts():
    assert list(gen_meshpatts(0)) == [MeshPatt(), MeshPatt((), [(0, 0)])]
    assert len(list(gen_meshpatts(1))) == 2 ** 4
    for i in range(2, 4):
        patt = Perm.random(i)
        len_i = list(gen_meshpatts(i, patt))
        assert (len(set(len_i))) == 2 ** ((i + 1) ** 2)
        for mpatt in len_i:
            assert mpatt.pattern == patt
    len_i = list(gen_meshpatts(2))
    assert len(set(len_i)) == (2 ** ((2 + 1) ** 2) * factorial(2))
    patt = tuple(Perm.random(3))
    len_i = list(gen_meshpatts(3, patt))
    assert (len(set(len_i))) == 2 ** ((3 + 1) ** 2)



class TrieNode(object):
    def __init__(self):
        self.down = {}
        self.value = None
        self.end = False

    def child(self, k):
        if k not in self.down:
            self.down[k] = TrieNode()
        return self.down[k]

    def height(self):
        return 1 + max([n.height() for n in self.down.values()] + [0])


class TrieMap(object):
    def __init__(self):
        self.root = TrieNode()
        self.cnt = 0

    def __contains__(self, key):
        cur = self.root
        for k in key:
            cur = cur.down.get(k, None)
            if cur is None:
                return False
        return cur.end

    def __setitem__(self, key, value):
        cur = self.root
        for k in key:
            cur = cur.child(k)
        cur.value = value
        if not cur.end:
            cur.end = True
            self.cnt += 1

    def __getitem__(self, key):
        cur = self.root
        for k in key:
            cur = cur.down.get(k, None)
            if cur is None:
                raise KeyError()
        if not cur.end:
            raise KeyError()
        return cur.value

    def height(self):
        return self.root.height()

    def __len__(self):
        return self.cnt



import numbers

from ._perm_set.finite.permset_static import PermSetStatic
from ._perm_set.permset_base import PermSetBase
from ._perm_set.unbounded.all.permset_all import PermSetAll
from ._perm_set.unbounded.described.avoiding.avoiding import Avoiding
from ._perm_set.unbounded.described.permset_described import PermSetDescribed
from .descriptors.basis import detect_basis_cls
from .descriptors.descriptor import Descriptor
from .descriptors.predicate import Predicate

__all__ = [
    "PermSet",
    "Av",
    "AvoidanceClass",
]


class PermSetMetaclass(type):
    def __instancecheck__(self, instance):
        return isinstance(instance, PermSetBase)

    def __subclasscheck__(self, subclass):
        return issubclass(subclass, PermSetBase)


class PermSet(object, metaclass=PermSetMetaclass):
    def __new__(cls, descriptor=None):
        if descriptor is None:
            return PermSetAll()
        elif isinstance(descriptor, numbers.Integral):
            # Descriptor is actually just a number
            return PermSetAll().of_length(descriptor)
        elif isinstance(descriptor, Descriptor):
            return cls._dispatch_described(descriptor)
        else:
            # Descriptor might just be a set of perms
            return PermSetStatic(descriptor)

    @classmethod
    def _dispatch_described(cls, descriptor):
        # Loop through all the described superclasses; e.g. Avoiding
        for described_superclass in PermSetDescribed.__subclasses__():
            # Check if descriptor's class is of that of the superclasses'
            # descriptor
            if isinstance(descriptor, described_superclass.DESCRIPTOR_CLASS):
                # The correct superclass has been found!
                # Try to find subclass specifically for this descriptor
                described_class = cls._find_described_class(
                    descriptor, PermSetDescribed
                )
                if described_class is None:
                    return described_superclass.DEFAULT_CLASS(descriptor)
                else:
                    return described_class(descriptor)
        # TODO: Something else?
        raise RuntimeError(
            "PermSet for descriptor {} not found".format(repr(descriptor))
        )

    @classmethod
    def _find_described_class(cls, descriptor, current_class):
        # TODO: Use metaclasses to track subclasses, rather than this mess
        if descriptor == current_class.DESCRIPTOR:
            return current_class
        else:
            for subclass in current_class.__subclasses__():
                described_class = cls._find_described_class(descriptor, subclass)
                if described_class is not None:
                    return described_class
            return None

    @classmethod
    def avoiding(cls, basis):
        BasisCls = detect_basis_cls(basis)
        return cls(BasisCls(basis))

    @classmethod
    def filtering(cls, predicate):
        return cls(Predicate(predicate))


#
# Syntactic sugar
#


class AvoidanceClassMetaclass(type):
    def __instancecheck__(self, instance):
        return isinstance(instance, Avoiding)

    def __subclasscheck__(self, subclass):
        return issubclass(subclass, Avoiding)


class Av(object, metaclass=AvoidanceClassMetaclass):
    def __new__(cls, basis=None):
        if basis is None or len(basis) == 0:
            return PermSetAll()
        return PermSet.avoiding(basis)


AvoidanceClass = Av



from permuta import Perm
from permuta.descriptors import Basis
from permuta.enumeration_strategies import all_enumeration_strategies, find_strategies
from permuta.enumeration_strategies.core_strategies import (
    RdCdCoreStrategy,
    RuCuCoreStrategy,
)
from permuta.enumeration_strategies.insertion_encodable import InsertionEncodingStrategy

ru = Perm((1, 2, 0, 3))
cu = Perm((2, 0, 1, 3))
rd = Perm((1, 3, 0, 2))
cd = Perm((2, 0, 3, 1))


def test_init_strategy():
    b1 = [Perm((0, 1, 2))]
    b2 = Basis([Perm((0, 1, 2, 3)), Perm((2, 0, 1))])
    for Strat in all_enumeration_strategies:
        Strat(b1).applies()
        Strat(b2).applies()


def test_insertion_encoding():
    strat = InsertionEncodingStrategy([Perm((0, 1, 2)), Perm((2, 0, 1))])
    assert strat.applies()
    strat = InsertionEncodingStrategy([Perm((0, 2, 1, 3))])
    assert not strat.applies()


def test_RuCu():
    assert not RuCuCoreStrategy([ru, cu, Perm((1, 4, 0, 2, 3))]).applies()
    assert not RuCuCoreStrategy([ru, Perm((0, 1, 2, 3))]).applies()
    assert RuCuCoreStrategy([ru, cu]).applies()
    assert RuCuCoreStrategy([ru, cu, Perm((0, 1, 2, 3, 4))]).applies()


def test_RdCd():
    assert not RdCdCoreStrategy([rd, cd, Perm((0, 1, 2, 3, 4))]).applies()
    assert not RdCdCoreStrategy([rd, Perm((0, 3, 2, 1))]).applies()
    assert RdCdCoreStrategy([rd, cd]).applies()
    assert RdCdCoreStrategy([rd, cd, Perm((0, 4, 1, 2, 3))]).applies()


def test_find_strategies():
    b1 = [Perm((0, 1, 2, 3, 4))]
    b2 = Basis([Perm((0, 1, 2, 3)), Perm((2, 0, 1))])
    assert len(find_strategies(b1, long_runnning=True)) == 0
    assert len(find_strategies(b1, long_runnning=False)) == 0
    assert len(find_strategies(b2)) > 0
    assert any(isinstance(s, InsertionEncodingStrategy) for s in find_strategies(b2))
    assert any(isinstance(s, RuCuCoreStrategy) for s in find_strategies([ru, cu]))



from abc import ABC, abstractmethod

from permuta.permutils.symmetry import all_symmetry_sets


class EnumerationStrategy(ABC):
    """Abstract class for a strategy to enumerate a permutation classes"""

    def __init__(self, basis):
        ABC.__init__(self)
        self._basis = frozenset(basis)

    @property
    def basis(self):
        return self._basis

    @classmethod
    def reference(cls):
        """A reference for the strategy."""
        raise NotImplementedError

    @abstractmethod
    def applies(self):
        """
        Return True if the strategy can be used for the basis
        """
        pass


class EnumerationStrategyWithSymmetry(EnumerationStrategy):
    """
    Abstract class for a strategy to enumerate a permutation classes.

    Each symmetry of the inputed basis is tested against the strategy.
    """

    def __init__(self, basis):
        super().__init__(basis)
        self._apply_basis = None

    @property
    def basis(self):
        """
        The symmetry of the inputed basis to which the strategy applies to.
        """
        if self._basis is None:
            self.applies()
        return self._basis

    def applies(self):
        """
        Check if the strategy applies to any symmetry.
        """
        for b in map(frozenset, all_symmetry_sets(self._basis)):
            if self._applies_to_symmetry(b):
                self._apply_basis = b
                return True
        return False

    @abstractmethod
    def _applies_to_symmetry(self, b):
        """
        Check if the strategy applies to this particular symmetry.
        """
        pass



"""
Enumeration strategies related to the insertion encoding.
"""

from permuta.enumeration_strategies.abstract_strategy import (
    EnumerationStrategyWithSymmetry,
)
from permuta.permutils.insertion_encodable import is_insertion_encodable_maximum


class InsertionEncodingStrategy(EnumerationStrategyWithSymmetry):
    def _applies_to_symmetry(self, b):
        return is_insertion_encodable_maximum(b)



from permuta import Av, MeshPatt, Perm
from permuta.permset import PermSetAll


def test_av_perm():
    p = Perm((0, 1))
    av = Av([p])
    for length in range(10):
        assert len(set(av.of_length(length))) == 1


def test_av_meshpatt():
    p = Perm((2, 0, 1))
    shading = ((2, 0), (2, 1), (2, 2), (2, 3))
    mp = MeshPatt(p, shading)
    av = Av([mp])
    enum = [1, 1, 2, 5, 15, 52, 203, 877]  # Bell numbers

    for (n, cnt) in enumerate(enum):
        inst = av.of_length(n)
        gen = list(inst)
        assert len(gen) == cnt


def test_av_empty():
    bases = [[], tuple(), set(), frozenset(), None]
    assert all(isinstance(Av(basis), PermSetAll) for basis in bases)



from . import avoiding_length_0
from .avoiding import Avoiding, AvoidingGeneric, AvoidingGenericSpecificLength

__all__ = [
    "avoiding_length_0",
    "Avoiding",
    "AvoidingGeneric",
    "AvoidingGenericSpecificLength",
]



import math
import sys
import time


class ProgressBar(object):
    @staticmethod
    def create(mx, mn=0):
        ProgressBar.mn = mn
        ProgressBar.mx = mx
        ProgressBar.at = mn
        ProgressBar.start = time.time()
        ProgressBar.last = 0
        ProgressBar.curw = 0
        # sys.stderr.write('\n')
        ProgressBar.progress(mn)

    @staticmethod
    def clear():
        sys.stderr.write("\r")
        sys.stderr.write(" " * ProgressBar.curw)
        sys.stderr.write("\r")

    @staticmethod
    def draw(fin=False):
        # sys.stderr.write('\033[1F')
        sys.stderr.write("\r")
        width = 50
        prog = (
            1
            if ProgressBar.mn == ProgressBar.mx
            else (
                float(ProgressBar.at - ProgressBar.mn)
                / (ProgressBar.mx - ProgressBar.mn)
            )
        )
        bars = int(round(prog * width))
        bars = max(0, min(width, bars))
        here = "%3d%% [%s%s] " % (round(prog * 100), "#" * bars, "-" * (width - bars))
        ProgressBar.curw = len(here)
        sys.stderr.write(here)
        elapsed = ProgressBar.last - ProgressBar.start
        # if elapsed >= 4 and prog > 0:
        show_time = None
        if fin:
            show_time = elapsed
        elif elapsed >= 2 and prog > 0:
            show_time = max(0, elapsed / prog - elapsed)
        if show_time is not None:
            h = math.floor(show_time / 60 / 60)
            show_time -= h * 60 * 60
            m = math.floor(show_time / 60)
            show_time -= m * 60
            s = math.floor(show_time)
            here = " %02d:%02d:%02d" % (h, m, s)
            sys.stderr.write(here)
            ProgressBar.curw += len(here)
        # sys.stderr.write('\n')

    @staticmethod
    def progress(prg=None, fin=False):
        if prg is not None:
            ProgressBar.at = prg
        else:
            ProgressBar.at = ProgressBar.at + 1
            prg = ProgressBar.at
        curt = time.time()
        if curt - ProgressBar.last < 0.5 and not fin:
            return
        ProgressBar.last = curt
        ProgressBar.draw(fin)

    @staticmethod
    def finish():
        ProgressBar.progress(ProgressBar.mx, fin=True)
        sys.stderr.write("\n")



import abc

ABC = abc.ABCMeta("ABC", (object,), {})


class Rotatable(ABC):
    def rotate(self, times=1):
        """Return self rotated 90 degrees to the right."""
        return self._rotate(times)

    rotate_right = rotate

    def rotate_left(self, times=1):
        """Return self rotated 90 degrees to the left."""
        return self._rotate(-times)

    def _rotate(self, times=1):
        """Return self rotated 90 times times degrees to the right."""
        times = times % 4
        if times == 0:
            return self
        elif times == 1:
            return self._rotate_right()
        elif times == 2:
            return self._rotate_180()
        else:
            return self._rotate_left()

    def _rotate_180(self):
        """Return self rotated 180 degrees."""
        return self._rotate_right()._rotate_right()

    def _rotate_left(self):
        """Return self rotated 90 degrees left."""
        return self._rotate_180()._rotate_right()
        pass

    @abc.abstractmethod
    def _rotate_right(self):
        """Return self rotated 90 degrees right."""
        pass



from itertools import combinations


def ordered_set_partitions(lst, parts, CACHE={}):
    key = tuple(parts)
    if key not in CACHE:
        CACHE[key] = list(helper(lst, parts))
    for partition in CACHE[key]:
        yield partition


def ordered_set_partitions_no_cache(lst, parts):
    for partition in list(helper(lst, parts)):
        yield partition


def helper(lst, parts):
    if not parts:
        yield []
    else:
        for comb in combinations(lst, parts[0]):
            new_lst = list(i for i in lst if i not in comb)
            for f in helper(new_lst, parts[1:]):
                yield [list(comb)] + f


def ordered_set_partitions_list(lst):
    for parts in partitions(len(lst)):
        if len(parts) == len(lst) or len(parts) == 1:
            continue
        for partition in ordered_set_partitions_no_cache(lst, parts):
            yield partition


def partitions(n, CACHE={}):
    if n not in CACHE:
        if n == 0:
            CACHE[n] = [[]]
        elif n == 1:
            CACHE[n] = [[1]]
        else:

            CACHE[n] = [[n]]
            for i in range(1, n):
                for part in partitions(n - i):
                    if part:
                        CACHE[n].append([i] + part)
    return CACHE[n]


def partitions_of_n_of_size_k(n, k):
    """possibly empty parts"""
    if n == 0 and k == 0:
        return [[]]
    elif k < 0:
        return []
    else:
        parts = []
        for i in range(n + 1):
            for part in partitions_of_n_of_size_k(n - i, k - 1):
                parts.append([i] + part)
    return parts


# internet versions

# def ordered_set_partitions(lst, parts):
#     return partitions(*parts)
#
# def partitions(*args):
#     def p(s, *args):
#         if not args: return [[]]
#         res = []
#         for c in combinations(s, args[0]):
#             s0 = [x for x in s if x not in c]
#             for r in p(s0, *args[1:]):
#                 res.append([c] + r)
#         return res
#     s = range(sum(args))
#     return p(s, *args)
#
# def partitions(*args):
#     def minus(s1, s2): return [x for x in s1 if x not in s2]
#     def p(s, *args):
#         if not args: return [[]]
#         return [[c] + r for c in combinations(s, args[0])
#                 for r in p(minus(s, c), *args[1:])]
#     return p(range(1, sum(args) + 1), *args)


# Raggis index

# def ordered_set_partitions(lst, parts):
#     if
#     result = [() for _ in range(len(parts))]
#     for partition in _helper(lst, parts, index):
#         yield partition
#
#
# def _helper(lst, parts, index):
#     if index >= len(parts):
#         yield
#
#
# def splits(lst, k):
#     for comb in combinations(lst, k):
#         yield comb, tuple(i for i in lst if i not in comb)



from .bisc import auto_bisc, bisc, create_bisc_input, read_bisc_file, write_bisc_files
from .bisc_subfunctions import (
    patterns_suffice_for_bad,
    patterns_suffice_for_good,
    run_clean_up,
    show_me,
    show_me_basis,
)

__all__ = [
    "auto_bisc",
    "bisc",
    "create_bisc_input",
    "patterns_suffice_for_bad",
    "patterns_suffice_for_good",
    "read_bisc_file",
    "write_bisc_files",
    "run_clean_up",
    "show_me",
    "show_me_basis",
]



from permuta.misc import cyclic_range, modulo_range


def test_cyclic_range():
    result = list(cyclic_range(5, 30, -20))
    expected = list(range(5, 30))
    expected.extend(range(-20, 5))
    assert result == expected


def test_modulo_range():
    result = list(modulo_range(7, 30))
    expected = list(range(7, 30))
    expected.extend(range(7))
    assert result == expected

    result = list(modulo_range(0, 30))
    expected = list(range(0, 30))
    assert result == expected



import numbers

INDEX_TYPE_ERROR_MESSAGE = "'{}' object is not a valid index"
INDEX_VALUE_ERROR_MESSAGE = "{} is not a valid index"


def index(obj, maximum=None):
    """Check if an object is an index and raise exceptions if it isn't.

    Args:
        obj: <object>
            The object to be checked.
        maximum: <numbers.Integral>
            The number the object is not to exceed if it is a number.

    Raises:
        TypeError:
            Object is not an integral number.
        ValueError:
            Object is an integral number, but not a valid index.

    Examples:
        >>> index(4)
        >>> index(0)
        >>> index(-3)
        Traceback (most recent call last):
            ...
        ValueError: -3 is not a valid index
        >>> index(21, 30)
        >>> index(0.3)
        Traceback (most recent call last):
            ...
        TypeError: '0.3' object is not a valid index
        >>> index(333, 30)
        Traceback (most recent call last):
            ...
        ValueError: 333 is not a valid index

    """
    if isinstance(obj, numbers.Integral):
        if obj >= 0 and (maximum is None or obj <= maximum):
            return  # Index is good
        else:
            raise ValueError(INDEX_VALUE_ERROR_MESSAGE.format(obj))
    else:
        raise TypeError(INDEX_TYPE_ERROR_MESSAGE.format(repr(obj)))



from ..permset_unbounded import PermSetUnbounded


class PermSetDescribed(PermSetUnbounded):
    """A base class for unbounded perm sets."""

    # TODO: Make things abstract properties as a solution?

    # TODO: These two attribute needs to be defined in all immediate subclasses
    DEFAULT_CLASS = None
    DESCRIPTOR_CLASS = None

    # TODO: This needs to be defined in classes for specific descriptors
    DESCRIPTOR = None

    def __init__(self, descriptor):
        self._descriptor = descriptor



import abc

ABC = abc.ABCMeta("ABC", (object,), {})


class Flippable(ABC):
    @abc.abstractmethod
    def flip_horizontal(self):
        """Return self flipped horizontally."""
        pass

    @abc.abstractmethod
    def flip_vertical(self):
        """Return self flipped vertically."""
        pass

    @abc.abstractmethod
    def flip_diagonal(self):
        """Return self flipped along the diagonal."""
        pass

    @abc.abstractmethod
    def flip_antidiagonal(self):
        """Return self flipped along the antidiagonal."""
        pass



from abc import abstractproperty, abstractstaticmethod

from permuta import Av, MeshPatt, Perm
from permuta.enumeration_strategies.abstract_strategy import (
    EnumerationStrategyWithSymmetry,
)

R_U = Perm((1, 2, 0, 3))
C_U = Perm((2, 0, 1, 3))
R_D = Perm((1, 3, 0, 2))
C_D = Perm((2, 0, 3, 1))


# Abstract Core Strategy


class CoreStrategy(EnumerationStrategyWithSymmetry):
    """
    Abstract class for a core related strategy.
    """

    @abstractproperty
    def patterns_needed():
        """
        Return the set of patterns that are needed for the strategy to be
        useful.
        """
        pass

    @abstractstaticmethod
    def is_valid_extension(patt):
        """
        Determine if the pattern satisfies the condition for strategy to apply.
        """
        pass

    def _applies_to_symmetry(self, b):
        """
        Check if the core strategy applies to the basis or any of its symmetry.

        INPUT:

        - `b`: a set of permutations.
        """
        assert isinstance(b, frozenset)
        perm_class = Av(b)
        patterns_are_contained = all(p not in perm_class for p in self.patterns_needed)
        extensions_are_valid = all(
            self.is_valid_extension(patt) for patt in b.difference(self.patterns_needed)
        )
        return patterns_are_contained and extensions_are_valid

    @classmethod
    def reference(cls):
        return (
            "Enumeration of Permutation Classes and Weighted Labelled "
            "Independent Sets: Corollary {}"
        ).format(cls.corr_number)

    @property
    @staticmethod
    def corr_number():
        """The number of the corollary in the that gives this strategy."""
        raise NotImplementedError


# Tool functions


def fstrip(perm):
    """
    Remove the leading 1 if the permutation is the sum of 1 + p.
    """
    if perm[0] == 0:
        return Perm.from_iterable(perm[1:])
    else:
        return perm


def bstrip(perm):
    """
    Remove the trailing n if the permutation is the sum of p + 1.
    """
    if perm[-1] == len(perm) - 1:
        return Perm.from_iterable(perm[:-1])
    else:
        return perm


def zero_plus_skewind(perm):
    """
    Return True if the permutation is of the form 1 + p where p is a
    skew-indecomposable permutations
    """
    return perm[0] == 0 and not fstrip(perm).skew_decomposable()


def zero_plus_sumind(perm):
    """
    Return True if the permutation is of the form 1 + p where p is a
    sum-indecomposable permutations
    """
    return perm[0] == 0 and not fstrip(perm).sum_decomposable()


def zero_plus_perm(perm):
    """
    Return True if the permutation starts with a zero.
    """
    return perm[0] == 0


def last_sum_component(p):
    """
    Return the last sum component of a permutation.
    """
    n = len(p)
    i = 1
    comp = set([p[n - i]])
    while comp != set(range(n - i, n)):
        i += 1
        comp.add(p[n - i])
    return Perm.to_standard(p[n - i : n])


def last_skew_component(p):
    """
    Return the last skew component of a permutation.
    """
    n = len(p)
    i = 1
    comp = set([p[n - i]])
    while comp != set(range(0, i)):
        i += 1
        comp.add(p[n - i])
    return Perm.to_standard(p[n - i : n])


# Core Strategies


class RuCuCoreStrategy(CoreStrategy):
    """
    This strategies uses independent set of the up-core graph to enumerate a
    class as inflation of an independent set.
    """

    patterns_needed = frozenset([R_U, C_U])
    is_valid_extension = staticmethod(zero_plus_skewind)
    corr_number = "4.3"


class RdCdCoreStrategy(CoreStrategy):
    """
    This strategies uses independent set of the down-core graph to enumerate a
    class as inflation of an independent set.
    """

    patterns_needed = frozenset([R_D, C_D])
    is_valid_extension = staticmethod(zero_plus_sumind)
    corr_number = "4.6"


class RuCuRdCdCoreStrategy(CoreStrategy):
    patterns_needed = frozenset([R_D, C_D, R_U, C_U])
    is_valid_extension = staticmethod(zero_plus_perm)
    corr_number = "5.4"


class RuCuCdCoreStrategy(CoreStrategy):
    patterns_needed = frozenset([R_U, C_U, C_D])
    is_valid_extension = staticmethod(zero_plus_skewind)
    corr_number = "6.3"


class RdCdCuCoreStrategy(CoreStrategy):
    patterns_needed = frozenset([R_D, C_D, C_U])
    corr_number = "7.4"

    @staticmethod
    def is_valid_extension(patt):
        return zero_plus_sumind(bstrip(patt))


class RdCuCoreStrategy(CoreStrategy):
    patterns_needed = frozenset([R_D, C_U])
    corr_number = "8.3"

    @staticmethod
    def is_valid_extension(patt):
        return zero_plus_skewind(patt) and zero_plus_sumind(bstrip(patt))


class Rd2134CoreStrategy(CoreStrategy):
    patterns_needed = frozenset([R_D, Perm((1, 0, 2, 3))])
    corr_number = "9.5"

    @staticmethod
    def is_valid_extension(patt):
        mp = MeshPatt(
            Perm((1, 0)), [(0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 1), (2, 2)]
        )
        last_comp = last_sum_component(fstrip(patt))
        return (
            patt[0] == 0
            and fstrip(patt).avoids(mp)
            and (last_comp not in Av([Perm((0, 1))]) or len(last_comp) == 1)
        )


class Ru2143CoreStrategy(CoreStrategy):
    patterns_needed = frozenset([R_U, Perm((1, 0, 3, 2))])
    corr_number = "10.5"

    @staticmethod
    def is_valid_extension(patt):
        mp = MeshPatt(
            Perm((0, 1)), [(0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 1), (2, 2)]
        )
        patt = fstrip(patt)
        return patt.avoids(mp) and last_skew_component(patt) not in Av([Perm((1, 0))])


core_strategies = [
    RuCuCoreStrategy,
    RdCdCoreStrategy,
    RuCuRdCdCoreStrategy,
    RuCuCdCoreStrategy,
    RdCdCuCoreStrategy,
    RdCuCoreStrategy,
    Rd2134CoreStrategy,
    Ru2143CoreStrategy,
]



import abc

from ..permset_base import PermSetBase


class PermSetFinite(PermSetBase):
    """Base class for all finite perm sets."""

    @abc.abstractmethod
    def random(self):
        # Return a random element from the range
        # Only possible due to the perm set being finite
        pass



class UnionFind(object):
    """A collection of distjoint sets."""

    def __init__(self, n=0):
        """Creates a collection of n disjoint unit sets."""
        self.p = [-1] * n
        self.leaders = set(i for i in range(n))

    def find(self, x):
        """Return the identifier of a representative element for the set
        containing the element with identifier x."""
        if self.p[x] < 0:
            return x
        self.p[x] = self.find(self.p[x])
        return self.p[x]

    def size(self, x):
        """Return the number of elements in the set containing the element with
        identifier x."""
        return -self.p[self.find(x)]

    def unite(self, x, y):
        """Unite the two sets containing the elements with identifiers x and y,
        respectively."""
        x = self.find(x)
        y = self.find(y)
        if x == y:
            return False
        if self.size(x) > self.size(y):
            x, y = y, x
        self.p[y] += self.p[x]
        self.p[x] = y
        self.leaders.remove(x)
        return True

    def add(self):
        """Add a unit set containing a new element to the collection, and
        return the identifier of the new element."""
        nid = len(self.p)
        self.p.append(nid)
        return nid



import pytest
from permuta import MeshPatt, Perm
from permuta.descriptors.basis import Basis, MeshBasis, detect_basis_cls


def test_detect_empty_basis_cls():
    assert detect_basis_cls(()) == Basis
    assert detect_basis_cls([]) == Basis


def test_detect_basis_cls_with_None_elmnt():
    # TypeError: 'NoneType' object is not iterable
    with pytest.raises(TypeError):
        detect_basis_cls(None)

    # ValueError: A basis can only contain Perms and MeshPatts.
    with pytest.raises(ValueError):
        detect_basis_cls([None])


def test_detect_basis_cls():
    p1 = Perm((0, 2, 1))
    p2 = Perm((2, 0, 1))
    perm_list = [p1, p2]

    assert detect_basis_cls(p1) == Basis
    assert detect_basis_cls(p2) == Basis
    assert detect_basis_cls(perm_list) == Basis
    assert detect_basis_cls(Basis) == Basis
    assert detect_basis_cls(Basis(perm_list)) == Basis


def test_detect_meshbasis_cls():
    p1 = Perm((0, 2, 1))
    p2 = Perm((2, 0, 1))
    shading = ((2, 0), (2, 1), (2, 2), (2, 3))
    mp1 = MeshPatt(p1, shading)
    mp2 = MeshPatt(p2, shading)
    meshpatt_list = [mp1, mp2]
    mixed_patt_list = [p1, mp2]

    assert detect_basis_cls(mp1) == MeshBasis
    assert detect_basis_cls(mp2) == MeshBasis
    assert detect_basis_cls(meshpatt_list) == MeshBasis
    assert detect_basis_cls(mixed_patt_list) == MeshBasis
    assert detect_basis_cls(MeshBasis) == MeshBasis
    assert detect_basis_cls(MeshBasis(meshpatt_list)) == MeshBasis


def test_meshbasis_of_perms():
    p1 = Perm((0, 2, 1))
    p2 = Perm((2, 0, 1))
    assert MeshBasis([p1, p2]) == MeshBasis([MeshPatt(p1, []), MeshPatt(p2, [])])

    shading = ((2, 0), (2, 1), (2, 2), (2, 3))
    mp1 = MeshPatt(p1, shading)
    mp2 = MeshPatt(p2, shading)
    assert MeshBasis([p1, p2]) == MeshBasis([p1, p2, mp1, mp2])

    # TypeError: Elements of a basis should all be of type(s) Perm
    with pytest.raises(TypeError):
        Basis([mp1, mp2])

    for elmnt in MeshBasis([p1, p2]):
        assert isinstance(elmnt, MeshPatt)


def test_basis_with_empty_perm():
    basis = Basis([Perm(), Perm((0,))])
    assert basis == Basis(Perm())


def test_meshbasis_with_empty_meshpatt():
    meshbasis = MeshBasis([MeshPatt(), MeshPatt(Perm((0,)), ())])
    assert meshbasis == MeshBasis(MeshPatt())


def test_meshbasis_with_empty_perm():
    meshbasis = MeshBasis([MeshPatt(), Perm()])
    assert meshbasis == MeshBasis(MeshPatt())



# Will return the set of polynomial types it intersects with
#   (W_++, W+-, W^-1 ++, L_2, L_2^R etc)
# 1: W++
# 2: W+-
# 3: W-+
# 4: W--
# 5: Winv++
# 6: Winv+-
# 7: Winv-+
# 8: Winv--
# 9: L2
# 10: L2inv
mem = {}


def types(perm):
    interset = mem.get(perm)
    if interset is None:
        interset = set([])
        for i in range(len(perm) + 1):
            part1 = perm[0:i]
            part2 = perm[i:]

            if is_incr(part1):
                if is_incr(part2):
                    interset.add(1)
                if is_decr(part2):
                    interset.add(2)

            if is_decr(part1):
                if is_incr(part2):
                    interset.add(3)
                if is_decr(part2):
                    interset.add(4)

        flipperm = perm.inverse()
        for i in range(len(perm) + 1):
            part1 = flipperm[0:i]
            part2 = flipperm[i:]

            if is_incr(part1):
                if is_incr(part2):
                    interset.add(5)
                if is_decr(part2):
                    interset.add(6)

            if is_decr(part1):
                if is_incr(part2):
                    interset.add(7)
                if is_decr(part2):
                    interset.add(8)
        if in_L2(perm):
            interset.add(9)
        if in_L2(perm.reverse()):
            interset.add(10)
    return interset


def is_decr(L):
    for i in range(len(L) - 1):
        if L[i] < L[i + 1]:
            return False
    return True


def is_incr(L):
    for i in range(len(L) - 1):
        if L[i] > L[i + 1]:
            return False
    return True


def in_L2(L):
    n = len(L)
    if n == 0 or n == 1:
        return True
    if L[-1] == n - 1:
        return in_L2(L[0 : n - 1])
    elif L[-1] == n - 2 and L[-2] == n - 1:
        return in_L2(L[0 : n - 2])
    else:
        return False


def is_polynomial(basis):
    overallinterset = set([])
    for perm in basis:
        overallinterset = overallinterset.union(types(perm))
        if len(overallinterset) == 10:
            return True
    return False


def is_non_polynomial(basis):
    return not is_polynomial(basis)



from permuta import Perm
from permuta.bisc.bisc_subfunctions import maximal_mesh_pattern_of_occurrence


def test_maximal_mesh_pattern_of_occurrence():
    assert maximal_mesh_pattern_of_occurrence(Perm((0,)), (0,)) == set(
        [(0, 0), (0, 1), (1, 0), (1, 1)]
    )
    assert maximal_mesh_pattern_of_occurrence(Perm((0, 1)), (0,)) == set(
        [(0, 0), (0, 1), (1, 0)]
    )
    assert maximal_mesh_pattern_of_occurrence(Perm((0, 1)), (1,)) == set(
        [(0, 1), (1, 0), (1, 1)]
    )
    assert maximal_mesh_pattern_of_occurrence(Perm((1, 0)), (0,)) == set(
        [(0, 0), (0, 1), (1, 1)]
    )
    assert maximal_mesh_pattern_of_occurrence(Perm((1, 0)), (1,)) == set(
        [(0, 0), (1, 0), (1, 1)]
    )



import collections

FloorAndCeiling = collections.namedtuple("FloorAndCeiling", ["floor", "ceiling"])


def left_floor_and_ceiling(iterable, default_floor=None, default_ceiling=None):
    """Find the left floor and ceiling indices of iterable.

    Define left_floor of an element in a sequence to be the index of the
    greatest smaller element to the left of said element, or default_floor
    if there is none. Similarly define default_ceiling. This function yields
    a tuple (left_floor, left_ceiling) for each element of iterable.

    Args:
        iterable: <iterable>
            An iterable of totally ordered unique elements.
        default_floor: <object>
        default_ceiling: <object>

    Yields: (int, int)
        The i-th yielded tuple is the left floor and ceiling indices of
        the i-th element of the iterable. The tuples are named tuples
        with floor and ceiling attributes.
    """
    # TODO: Define behaviour for duplicate elements
    dq = collections.deque()
    smallest = None
    biggest = None
    index = 0
    for element in iterable:
        if index == 0:
            dq.append((element, index))
            smallest = element
            biggest = element
            yield FloorAndCeiling(default_floor, default_ceiling)
        else:
            if element <= smallest:
                # Rotate until smallest element is at front
                while dq[0][0] != smallest:
                    dq.rotate(-1)
                yield FloorAndCeiling(default_floor, dq[0][1])
                dq.appendleft((element, index))
                smallest = element
            elif element >= biggest:
                # Rotate until biggest element is at end
                while dq[-1][0] != biggest:
                    dq.rotate(-1)
                yield FloorAndCeiling(dq[-1][1], default_ceiling)
                dq.append((element, index))
                biggest = element
            else:
                while not dq[-1][0] <= element <= dq[0][0]:
                    dq.rotate()
                yield FloorAndCeiling(dq[-1][1], dq[0][1])
                dq.appendleft((element, index))
        index += 1


def right_floor_and_ceiling(iterable, default_floor=None, default_ceiling=None):
    """The right counterpart of left_floor_and_ceiling."""
    # TODO: Implement nicely
    result = left_floor_and_ceiling(
        reversed(list(iterable)), default_floor, default_ceiling
    )
    for fac in reversed(list(result)):
        yield fac
